
var Formstone = window.Formstone = (function ($, window, document, undefined) {

	"use strict";

	var Core = function() {
			this.Version = '@version';
			this.Plugins = {};

			this.DontConflict   = false;
			this.Conflicts      = {
				fn: {}
			};
			this.ResizeHandlers = [];
			this.RAFHandlers    = [];

			this.window               = window;
			this.$window              = $(window);
			this.document             = document;
			this.$document            = $(document);
			this.$body                = null;

			this.windowWidth          = 0;
			this.windowHeight         = 0;
			this.fallbackWidth        = 1024;
			this.fallbackHeight       = 768;
			this.userAgent            = window.navigator.userAgent || window.navigator.vendor || window.opera;
			this.isFirefox            = /Firefox/i.test(this.userAgent);
			this.isChrome             = /Chrome/i.test(this.userAgent);
			this.isSafari             = /Safari/i.test(this.userAgent) && !this.isChrome;
			this.isMobile             = /Android|webOS|iPhone|iPad|iPod|BlackBerry/i.test( this.userAgent );
			this.isIEMobile           = /IEMobile/i.test(this.userAgent);
			this.isFirefoxMobile      = (this.isFirefox && this.isMobile);
			this.transform            = null;
			this.transition           = null;

			this.support = {
				file          : !!(window.File && window.FileList && window.FileReader),
				history       : !!(window.history && window.history.pushState && window.history.replaceState),
				matchMedia    : !!(window.matchMedia || window.msMatchMedia),
				pointer       : !!(window.PointerEvent),
				raf           : !!(window.requestAnimationFrame && window.cancelAnimationFrame),
				touch         : !!(("ontouchstart" in window) || window.DocumentTouch && document instanceof window.DocumentTouch),
				transition    : false,
				transform     : false
			};
		},

		Functions = {
			
			killEvent: function(e, immediate) {
				try {
					e.preventDefault();
					e.stopPropagation();

					if (immediate) {
						e.stopImmediatePropagation();
					}
				} catch(error) {
					
				}
			},

			startTimer: function(timer, time, callback, interval) {
				Functions.clearTimer(timer);

				return (interval) ? setInterval(callback, time) : setTimeout(callback, time);
			},

			clearTimer: function(timer, interval) {
				if (timer) {
					if (interval) {
						clearInterval(timer);
					} else {
						clearTimeout(timer);
					}

					timer = null;
				}
			},


			sortAsc: function(a, b) {
				return (parseInt(a, 10) - parseInt(b, 10));
			},


			sortDesc: function(a, b) {
				return (parseInt(b, 10) - parseInt(a, 10));
			},

			
			decodeEntities: function(string) {
				var el = Formstone.document.createElement("textarea");
				el.innerHTML = string;

				return el.value;
			},

			parseQueryString: function(url) {
				var params = {},
					parts = url.slice( url.indexOf("?") + 1 ).split("&");

				for (var i = 0; i < parts.length; i++) {
					var part = parts[i].split("=");
					params[ part[0] ] = part[1];
				}

				return params;
		    }
		},

		Formstone = new Core(),

		$Ready = $.Deferred(),

		Classes = {
			base                 : "{ns}",
			element              : "{ns}-element"
		},

		Events = {
			namespace            : ".{ns}",
			beforeUnload         : "beforeunload.{ns}",
			blur                 : "blur.{ns}",
			change               : "change.{ns}",
			click                : "click.{ns}",
			dblClick             : "dblclick.{ns}",
			drag                 : "drag.{ns}",
			dragEnd              : "dragend.{ns}",
			dragEnter            : "dragenter.{ns}",
			dragLeave            : "dragleave.{ns}",
			dragOver             : "dragover.{ns}",
			dragStart            : "dragstart.{ns}",
			drop                 : "drop.{ns}",
			error                : "error.{ns}",
			focus                : "focus.{ns}",
			focusIn              : "focusin.{ns}",
			focusOut             : "focusout.{ns}",
			input                : "input.{ns}",
			keyDown              : "keydown.{ns}",
			keyPress             : "keypress.{ns}",
			keyUp                : "keyup.{ns}",
			load                 : "load.{ns}",
			mouseDown            : "mousedown.{ns}",
			mouseEnter           : "mouseenter.{ns}",
			mouseLeave           : "mouseleave.{ns}",
			mouseMove            : "mousemove.{ns}",
			mouseOut             : "mouseout.{ns}",
			mouseOver            : "mouseover.{ns}",
			mouseUp              : "mouseup.{ns}",
			panStart             : "panstart.{ns}",
			pan                  : "pan.{ns}",
			panEnd               : "panend.{ns}",
			resize               : "resize.{ns}",
			scaleStart           : "scalestart.{ns}",
			scaleEnd             : "scaleend.{ns}",
			scale                : "scale.{ns}",
			scroll               : "scroll.{ns}",
			select               : "select.{ns}",
			swipe                : "swipe.{ns}",
			touchCancel          : "touchcancel.{ns}",
			touchEnd             : "touchend.{ns}",
			touchLeave           : "touchleave.{ns}",
			touchMove            : "touchmove.{ns}",
			touchStart           : "touchstart.{ns}"
		};

	Core.prototype.NoConflict = function() {
		Formstone.DontConflict = true;

		for (var i in Formstone.Plugins) {
			if (Formstone.Plugins.hasOwnProperty(i)) {
				$[i]    = Formstone.Conflicts[i];
				$.fn[i] = Formstone.Conflicts.fn[i];
			}
		}
	};

	Core.prototype.Plugin = function(namespace, settings) {
		Formstone.Plugins[namespace] = (function(namespace, settings) {

			var namespaceDash  = "fs-" + namespace,
				namespaceDot   = "fs." + namespace,
				namespaceClean = "fs"  + namespace.replace(/(^|\s)([a-z])/g, function(m, p1, p2) { return p1 + p2.toUpperCase(); });

			function initialize(options) {
				
				var hasOptions = $.type(options) === "object",
					$targets = this,
					$postTargets = $(),
					$element,
					i,
					count;

				options = $.extend(true, {}, settings.defaults || {}, (hasOptions ? options : {}));

				for (i = 0, count = $targets.length; i < count; i++) {
					$element = $targets.eq(i);
					
					if (!getData($element)) {

						var guid    = "__" + settings.guid++,
							rawGuid = settings.classes.raw.base + guid,
							locals  = $element.data(namespace + "-options"),
							data    = $.extend(true, {
								$el     : $element,
								guid    : guid,
								rawGuid : rawGuid,
								dotGuid : "." + rawGuid
							}, options, ($.type(locals) === "object" ? locals : {}) );

						$element.addClass(settings.classes.raw.element)
						        .data(namespaceClean, data);

						settings.methods._construct.apply($element, [ data ].concat(Array.prototype.slice.call(arguments, (hasOptions ? 1 : 0) )));

						$postTargets = $postTargets.add($element);
					}

				}

				for (i = 0, count = $postTargets.length; i < count; i++) {
					$element = $postTargets.eq(i);
					settings.methods._postConstruct.apply($element, [ getData($element) ]);
				}

				return $targets;
			}

			function destroy(data) {
				settings.functions.iterate.apply(this, [ settings.methods._destruct ].concat(Array.prototype.slice.call(arguments, 1)));

				this.removeClass(settings.classes.raw.element)
					.removeData(namespaceClean);
			}

			function getData($element) {
				return $element.data(namespaceClean);
			}

			function delegateWidget(method) {

				if (this instanceof $) {

					var _method = settings.methods[method];

					if ($.type(method) === "object" || !method) {

						return initialize.apply(this, arguments);
					} else if (_method && method.indexOf("_") !== 0) {

						return settings.functions.iterate.apply(this, [ _method ].concat(Array.prototype.slice.call(arguments, 1)));
					}

					return this;
				}
			}

			function delegateUtility(method) {

				var _method = settings.utilities[method] || settings.utilities._initialize || false;

				if (_method) {

					return _method.apply(window, Array.prototype.slice.call(arguments, ($.type(method) === "object" ? 0 : 1) ));
				}
			}

			function defaults(options) {
				settings.defaults = $.extend(true, settings.defaults, options || {});
			}

			function iterate(fn) {
				var $targets = this;

				for (var i = 0, count = $targets.length; i < count; i++) {
					var $element = $targets.eq(i),
						data = getData($element) || {};

					if ($.type(data.$el) !== "undefined") {
						fn.apply($element, [ data ].concat(Array.prototype.slice.call(arguments, 1)));
					}
				}

				return $targets;
			}

			settings.initialized = false;
			settings.priority    = settings.priority || 10;
			settings.classes   = namespaceProperties("classes", namespaceDash, Classes, settings.classes);
			settings.events    = namespaceProperties("events",  namespace,     Events,  settings.events);

			settings.functions = $.extend({
				getData    : getData,
				iterate    : iterate
			}, Functions, settings.functions);

			settings.methods = $.extend(true, {

				_setup         : $.noop,    
				_construct     : $.noop,    
				_postConstruct : $.noop,    
				_destruct      : $.noop,    
				_resize        : false,     

				destroy        : destroy

			}, settings.methods);


			settings.utilities = $.extend(true, {

				_initialize    : false,    
				_delegate      : false,    

				defaults       : defaults

			}, settings.utilities);


			if (settings.widget) {

				Formstone.Conflicts.fn[namespace] = $.fn[namespace];

				$.fn[namespaceClean] = delegateWidget;

				if (!Formstone.DontConflict) {

					$.fn[namespace] = $.fn[namespaceClean];
				}
			}

				Formstone.Conflicts[namespace] = $[namespace];

				$[namespaceClean] = settings.utilities._delegate || delegateUtility;

				if (!Formstone.DontConflict) {

					$[namespace] = $[namespaceClean];
				}

			settings.namespace      = namespace;
			settings.namespaceClean = namespaceClean;

			settings.guid = 0;

			if (settings.methods._resize) {
				Formstone.ResizeHandlers.push({
					namespace: namespace,
					priority: settings.priority,
					callback: settings.methods._resize
				});

				Formstone.ResizeHandlers.sort(sortPriority);
			}

			if (settings.methods._raf) {
				Formstone.RAFHandlers.push({
					namespace: namespace,
					priority: settings.priority,
					callback: settings.methods._raf
				});

				Formstone.RAFHandlers.sort(sortPriority);
			}

			return settings;
		})(namespace, settings);

		$Ready.then(function() {
			setupPlugin(namespace);
		});

		return Formstone.Plugins[namespace];
	};

	function setupPlugin(namespace) {
		if (!Formstone.Plugins[namespace].initialized) {
			Formstone.Plugins[namespace].methods._setup.call(document);
			Formstone.Plugins[namespace].initialized = true;
		}
	}

	function namespaceProperties(type, namespace, globalProps, customProps) {
		var _props = {
				raw: {}
			},
			i;

		customProps = customProps || {};

		for (i in customProps) {
			if (customProps.hasOwnProperty(i)) {
				if (type === "classes") {

					_props.raw[ customProps[i] ] = namespace + "-" + customProps[i];
					_props[ customProps[i] ]     = "." + namespace + "-" + customProps[i];
				} else {
					_props.raw[ i ] = customProps[i];
					_props[ i ]     = customProps[i] + "." + namespace;
				}
			}
		}

		for (i in globalProps) {
			if (globalProps.hasOwnProperty(i)) {
				if (type === "classes") {

					_props.raw[ i ] = globalProps[i].replace(/{ns}/g, namespace);
					_props[ i ]     = globalProps[i].replace(/{ns}/g, "." + namespace);
				} else {
					_props.raw[ i ] = globalProps[i].replace(/.{ns}/g, "");
					_props[ i ]     = globalProps[i].replace(/{ns}/g, namespace);
				}
			}
		}

		return _props;
	}


	function setTransitionInformation() {
		var transitionEvents = {
				"WebkitTransition"    : "webkitTransitionEnd",
				"MozTransition"       : "transitionend",
				"OTransition"         : "otransitionend",
				"transition"          : "transitionend"
			},
			transitionProperties = [
				"transition",
				"-webkit-transition"
			],
			transformProperties = {
				'transform'          : 'transform',
				'MozTransform'       : '-moz-transform',
				'OTransform'         : '-o-transform',
				'msTransform'        : '-ms-transform',
				'webkitTransform'    : '-webkit-transform'
			},
			transitionEvent       = "transitionend",
			transitionProperty    = "",
			transformProperty     = "",
			test                  = document.createElement("div"),
			i;


		for (i in transitionEvents) {
			if (transitionEvents.hasOwnProperty(i) && i in test.style) {
				transitionEvent = transitionEvents[i];
				Formstone.support.transition = true;
				break;
			}
		}

		Events.transitionEnd = transitionEvent + ".{ns}";

		for (i in transitionProperties) {
			if (transitionProperties.hasOwnProperty(i) && transitionProperties[i] in test.style) {
				transitionProperty = transitionProperties[i];
				break;
			}
		}

		Formstone.transition = transitionProperty;

		for (i in transformProperties) {
			if (transformProperties.hasOwnProperty(i) && transformProperties[i] in test.style) {
				Formstone.support.transform = true;
				transformProperty = transformProperties[i];
				break;
			}
		}

		Formstone.transform = transformProperty;
	}

	var ResizeTimer = null,
		Debounce = 20;

	function onWindowResize() {
		Formstone.windowWidth  = Formstone.$window.width();
		Formstone.windowHeight = Formstone.$window.height();

		ResizeTimer = Functions.startTimer(ResizeTimer, Debounce, handleWindowResize);
	}

	function handleWindowResize() {
		for (var i in Formstone.ResizeHandlers) {
			if (Formstone.ResizeHandlers.hasOwnProperty(i)) {
				Formstone.ResizeHandlers[i].callback.call(window, Formstone.windowWidth, Formstone.windowHeight);
			}
		}
	}

	Formstone.$window.on("resize.fs", onWindowResize);
	onWindowResize();

	function handleRAF() {
		if (Formstone.support.raf) {
			Formstone.window.requestAnimationFrame(handleRAF);

			for (var i in Formstone.RAFHandlers) {
				if (Formstone.RAFHandlers.hasOwnProperty(i)) {
					Formstone.RAFHandlers[i].callback.call(window);
				}
			}
		}
	}

	handleRAF();

	function sortPriority(a, b) {
		return (parseInt(a.priority) - parseInt(b.priority));
	}

	$(function() {
		Formstone.$body = $("body");

		$Ready.resolve();

		Formstone.support.nativeMatchMedia = Formstone.support.matchMedia && !$("html").hasClass("no-matchmedia");
	});

	Events.clickTouchStart = Events.click + " " + Events.touchStart;

	setTransitionInformation();

	return Formstone;

})(jQuery, window, document);